// Copyright 2023 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package configfsi

import (
	"strconv"
	"strings"
)

// Kstrtouint returns the unsigned integer represented in data, following the same grammar
// allowances as Linux, i.e., data may have a single trailing newline.
func Kstrtouint(data []byte, base, bits int) (uint64, error) {
	return strconv.ParseUint(strings.TrimRight(string(data), "\n"), base, bits)
}
